@@ -7,6 +7,7 @@ from sqlalchemy.orm import Session
 import datetime
 from copy import copy
 from pyg_base import logger
+from functools import partial
 
 _id = '_id'
 _doc = 'doc'
@@ -117,13 +118,9 @@ def create_schema(engine, schema):
             engine.execute(sa.schema.CreateSchema(schema))
             logger.info('creating schema: %s'%schema)
     return schema
-    
-def get_engine(*pairs, **connection):    
-    """
-    returns a sqlalchemy engine object
-    accepts either *pairs: 'driver={ODBC Driver 17 for SQL Server}'
-    or keyword arguments that look like driver = 'ODBC Driver 17 for SQL Server'    
-    """
+
+@cache
+def _get_engine(*pairs, **connection):    
     connection = _pairs2connection(*pairs, **connection)
     server = get_server(connection.pop('server', None))
     connection['driver'] = get_driver(connection.pop('driver', None))
@@ -139,6 +136,14 @@ def get_engine(*pairs, **connection):
         create_database(cstr)
         e = sa.create_engine(cstr)       
     return e
+
+def get_engine(*pairs, **connection):
+    """
+    returns a sqlalchemy engine object
+    accepts either *pairs: 'driver={ODBC Driver 17 for SQL Server}'
+    or keyword arguments that look like driver = 'ODBC Driver 17 for SQL Server'    
+    """
+    return _get_engine(*pairs, **connection)
     
 _types = {str: String, 'str' : String, 
           int : Integer, 'int' : Integer,
@@ -151,6 +156,12 @@ _types = {str: String, 'str' : String,
 
 _orders = {1 : asc, True: asc, 'asc': asc, asc : asc, -1: desc, False: desc, 'desc': desc, desc: desc}
 
+@cache
+def _get_table(table_name, schema, db, server):
+    e = _get_engine(server = server, db = db, schema = schema)
+    meta = MetaData()
+    return Table(table_name, meta, autoload_with = e, schema = schema)
+
 def sql_table(table, db = None, non_null = None, nullable = None, _id = None, schema = None, server = None, reader = None, writer = None, pk = None, doc = None, mode = None):
     """
     Creates a sql table. Can also be used to simply read table from the db
@@ -202,7 +213,16 @@ def sql_table(table, db = None, non_null = None, nullable = None, _id = None, sc
             table = values[1]
         elif len(values)>2:
             raise ValueError('not sure how to translate this %s into a db.table format'%table)
-    e = get_engine(server = server, db = db, schema = schema)
+    elif isinstance(table, partial): # support for using the partial rather than the actual table
+        db = table.keywords.get('db') if db is None else db
+        schema = table.keywords.get('schema') if schema is None else schema
+        server = table.keywords.get('server') if server is None else server
+        pk = table.keywords.get('pk') if pk is None else pk
+        writer = table.keywords.get('writer') if writer is None else writer
+        doc = table.keywords.get('doc') if doc is None else doc
+        table = table.keywords['table']
+        
+    e = _get_engine(server = server, db = db, schema = schema)
     
     non_null = non_null or {}
     nullable = nullable or {}
@@ -229,9 +249,12 @@ def sql_table(table, db = None, non_null = None, nullable = None, _id = None, sc
     schema = create_schema(e, _schema(schema))
     if doc is True:
         doc = _doc
-    meta = MetaData()
-    i = sa.inspect(e)
-    if not i.has_table(table_name, schema = schema):
+    try:
+        tbl = _get_table(table_name, schema, db, server) ## by default we grab the existing table
+    except sa.exc.NoSuchTableError:        
+        meta = MetaData()
+        # i = sa.inspect(e)
+        # if not i.has_table(table_name, schema = schema):
         cols = []
         if isinstance(table, sa.sql.schema.Table):
             for col in table.columns:
@@ -262,18 +285,21 @@ def sql_table(table, db = None, non_null = None, nullable = None, _id = None, sc
         logger.info('creating table: %s.%s.%s%s'%(db, schema, table_name, [col.name for col in cols]))
         tbl = Table(table_name, meta, *cols, schema = schema)
         meta.create_all(e)
-    else:
-        # logger.info('schema :%s'%schema)
-        tbl = Table(table_name, meta, autoload_with = e, schema = schema)
-        cols = tbl.columns
-        non_nulls = [Column(k, _types.get(t, t), nullable = False) for k, t in pks.items()]
-        if non_nulls is not None:
-            for key in non_nulls:
-                if key.name not in cols.keys():
-                    raise ValueError('column %s does not exist in %s.%s'%(key, db, table_name))
-                elif cols[key.name].nullable is True:
-                    raise ValueError('WARNING: You defined %s as a primary but it is nullable in %s.%s'%(key, db, table_name))
-    res = sql_cursor(table = tbl, db = db, server = server, engine = e, spec = None, selection = None, reader = reader, writer = writer, pk = pk, doc = doc)
+    # else:
+    #     pass
+    #     # logger.info('schema :%s'%schema)
+    #     tbl = Table(table_name, meta, autoload_with = e, schema = schema)
+        # cols = tbl.columns
+        # non_nulls = [Column(k, _types.get(t, t), nullable = False) for k, t in pks.items()]
+        # if non_nulls is not None:
+        #     for key in non_nulls:
+        #         if key.name not in cols.keys():
+        #             raise ValueError('column %s does not exist in %s.%s'%(key, db, table_name))
+        #         elif cols[key.name].nullable is True:
+        #             raise ValueError('WARNING: You defined %s as a primary but it is nullable in %s.%s'%(key, db, table_name))
+    res = sql_cursor(table = tbl, schema = schema, db = db, server = server, engine = e, 
+                     spec = None, selection = None, reader = reader, writer = writer, 
+                     pk = pk, doc = doc)
     return res
 
 class sql_cursor(object):
@@ -488,7 +514,7 @@ class sql_cursor(object):
     >>> assert list(read_from_db.salary.values) == [100, 200, 300]
     >>> assert list(read_from_file.values) == [100, 200, 300]
     """
-    def __init__(self, table, db = None, engine = None, server = None, spec = None, selection = None, order = None, reader = None, writer = None, pk = None, doc = None, **_):
+    def __init__(self, table, schema = None, db = None, engine = None, server = None, spec = None, selection = None, order = None, reader = None, writer = None, pk = None, doc = None, **_):
         """
         Parameters
         ----------
@@ -496,6 +522,8 @@ class sql_cursor(object):
             Our table
         db : string, optional
             Name of the db where table is.
+        schema : string, optional
+            Name of the db schema for table.
         engine : sa,Engine, optional
             The sqlalchemy engine
         server : str , optional
@@ -515,7 +543,7 @@ class sql_cursor(object):
             
         """
         if is_str(table):
-            table = sql_table(table = table, db = db, server = server)
+            table = sql_table(table = table, schema = schema, db = db, server = server)
             
         if isinstance(table, sql_cursor):
             db = table.db if db is None else db
@@ -523,6 +551,7 @@ class sql_cursor(object):
             server = table.server if server is None else server
             spec = table.spec if spec is None else spec
             selection = table.selection if selection is None else selection
+            schema = table.schema if schema is None else schema
             order = table.order if order is None else order
             reader = table.reader if reader is None else reader
             writer = table.writer if writer is None else writer
@@ -531,6 +560,7 @@ class sql_cursor(object):
             table = table.table
     
         self.table = table
+        self.schema = schema
         self.db = db
         self.server = get_server(server)
         self.engine = engine or get_engine(db = self.db, server = self.server)
@@ -544,13 +574,7 @@ class sql_cursor(object):
     
     def copy(self):
         return type(self)(self)
-    
-    @property
-    def schema(self):
-        """
-        table schema
-        """
-        return self.table.schema        
+
     
     @property
     def _ids(self):
@@ -1261,12 +1285,13 @@ class sql_cursor(object):
     def __repr__(self):
         statement = self.statement()
         params = statement.compile().params
-        text = str(statement).replace(self.table.name+'.','')
+        prefix = '%s.%s'%(self.schema, self.table.name) if self.schema else self.table.name
+        text = str(statement).replace(prefix + '.','')
         for k,v in params.items():
             text = text.replace(':'+k, '"%s"'%v if is_str(v) else dt2str(v) if is_date(v) else str(v))
             
-        res = 'sql_cursor: %(db)s.%(table)s%(pk)s %(doc)s %(w)s\n%(statement)s\n%(n)i records'%dict(db = self.db, table = self.table.name, doc = 'DOCSTORE[%s]'%self.doc if self.doc else '', 
-                                                                                              w = 'writer = %s'%self.writer if is_str(self.writer) else '',
+        res = 'sql_cursor: %(db)s.%(table)s%(pk)s %(doc)s %(w)s\n%(statement)s\n%(n)i records'%dict(db = self.db, table = prefix, doc = 'DOCSTORE[%s]'%self.doc if self.doc else '', 
+                                                                                              w = '\nwriter: %s\n'%self.writer if is_str(self.writer) else '',
                                                                                               pk = self._pk if self._pk else '', 
                                                                                               n = len(self), 
                                                                                               statement = text)
@@ -1282,7 +1307,9 @@ class sql_cursor(object):
         else:        
             schema = _archived + (self.schema or '')
             # logger.info('archived schema: %s'%schema)
-            res = sql_table(table = self.table, db = self.db, non_null = dict(deleted = datetime.datetime), 
+            res = sql_table(table = self.table, 
+                            db = self.db, 
+                            non_null = dict(deleted = datetime.datetime), 
                             server = self.server, 
                             pk = self.pk, 
                             doc = self.doc, 
